<!DOCTYPE html><html lang="en"><head><title>{% block title %}{% endblock %} - Namer</title><link href="{{ url_for('.static', filename='assets/bundle.min.css') }}" rel="stylesheet"/><link rel="icon" type="image/png" href="{{ url_for('.static', filename='favicon.png') }}"/></head><body><div class="container">{% block content %}{% endblock %}</div><script src="{{ url_for('.static', filename='assets/bundle.min.js') }}"></script></body></html>