{% if data %} <table class="table table-sm text-start align-middle"><thead><tr><th scope="col">File Name</th><th scope="col">Extension</th><th scope="col">Size</th><th scope="col"></th></tr></thead><tbody> {% for file in data %} <tr><td class="col">{{ file['name'] }}</td><td class="col-1">{{ file['ext'] }}</td><td class="col-1">{{ file['size'] }}</td><td class="col-2 text-end"><button type="button" class="btn btn-primary match" data-bs-toggle="modal" data-bs-target="#searchForm" data-file="{{ file['file'] }}"><i class="bi bi-search"></i></button> {% if config.allow_delete_files %} <button type="button" class="btn btn-danger delete" data-bs-toggle="modal" data-bs-target="#deleteConfirm" data-file="{{ file['file'] }}"><i class="bi bi-trash"></i></button> {% endif %} </td></tr> {% endfor %} </tbody></table> {% else %} <div class="col m-1"><span>No files found</span></div> {% endif %}